@@ -60,14 +60,17 @@ from rich.progress import (
     Progress,
     track,
 )
+from rich.traceback import install
 from concurrent.futures import ThreadPoolExecutor
 from concurrent.futures import as_completed
 from functools import partial
 from signal import signal, SIGPIPE, SIGINT, SIG_DFL
 from .__version__ import __version__
+from .biosample import BioSample, KEYS, get_bs2assemid
 
 signal(SIGPIPE, SIG_DFL)
 signal(SIGINT, SIG_DFL)
+install(show_locals=True)
 
 NOT_WHITESPACE = re.compile(r"[^\s]")
 OUTPUTS = ("fna", "ffn", "gff", "gbk", "faa")
@@ -104,12 +107,27 @@ class DownloadEdirect(argparse.Action):
         )
         logger.info("This script will find ~/edirect by default.")
         logger.info(
-            "Alternatively, set [green]$EDIRECT[/green] environment " "variable.",
+            "Alternatively, set [green]$EDIRECT[/green] environment variable.",
             extra={"markup": True},
         )
         parser.exit()
 
 
+def extract_attribute(bs: BioSample, attr: dict):
+    try:
+        name = attr['attribute_name']
+    except TypeError:
+        pass
+    else:
+        if name in KEYS:
+            setattr(bs, name, attr['content'])
+        elif name in ('sample name', 'strain', 'isolate'):
+            if attr['content'] == 'not applicable':
+                pass
+            elif bs.isolate == '-':
+                setattr(bs, 'isolate', attr['content'])
+
+
 def dl_file(url, out_file):
     logger = logging.getLogger(__name__)
     logger.debug("Downloading file {} to {}".format(url, out_file))
@@ -272,7 +290,7 @@ def run_argparse():
     for p in p_assemlinks:
         p.add_argument(
             "--function",
-            help="check counts, download metadata," " or genomes. [check]",
+            help="check counts, download metadata, or genomes. [check]",
             type=str,
             default=["check"],
             choices=["check", "metadata", "genomes"],
@@ -301,7 +319,7 @@ def run_argparse():
         )
         p.add_argument(
             "--keepmulti",
-            help="By default, genomes from large multi-isolate" "studies are removed.",
+            help="By default, genomes from large multi-isolate studies are removed.",
             action="store_true",
             default=False,
         )
@@ -335,6 +353,10 @@ def run_argparse():
             choices=["fna", "ffn", "gff", "gbk", "faa", "all"],
             nargs="+",
         )
+        p.add_argument(
+            "--no-biosample", help="Turn off biosample data.", action="store_true",
+            dest="biosample"
+        )
         p.add_argument("--edirect", help="Path to edirect directory.", type=edirect_dir)
         p.add_argument(
             "--debug", help="Turn on debugging messages.", action="store_true"
@@ -407,7 +429,7 @@ def validate_inputs(args):
         logger.debug("No edirect dir found.")
         edirect_dir = ""
 
-    utils = ["esearch", "efetch", "xtract", "efilter", "elink", "epost"]
+    utils = ["esearch", "efetch", "xtract", "efilter", "elink", "epost", "xml2json"]
     exes = {}
     for util in utils:
         if shutil.which(util):
@@ -435,7 +457,7 @@ def search_query(esearch, efilter, qtype, query, typestrain):
     elif qtype == "ID":
         query = f"txid{query}[ORGN]"
 
-    query = query + " AND latest [PROP] NOT " "partial-genome-representation [PROP]"
+    query = query + " AND latest [PROP] NOT partial-genome-representation [PROP]"
 
     if typestrain:
         query = query + ' AND "from type" [PROP]'
@@ -466,6 +488,8 @@ def get_assem_links(epost, infile, qtype):
     command = [epost, "-db", "assembly", "-format", f"{qtype}"]
 
     # command = ' '.join(command)
+    logger.info('Getting assembly links. This may take a while.')
+    logger.debug(' '.join(command))
     try:
         output = subprocess.run(
             command,
@@ -809,6 +833,7 @@ def extract_metadata(
             annotation_type = "refseq"
             line = []
             skip = False
+            biosample = ''
             # Standard keys; only str/int return
             for json_key in json_keys:
                 try:
@@ -821,6 +846,8 @@ def extract_metadata(
                         line.append(str(item))
                     else:
                         line.append("-")
+                    if json_key == "biosampleaccn":
+                        biosample = item
                     if json_key == "fromtype":
                         if not item and typestrain:
                             skip = True
@@ -970,7 +997,7 @@ def extract_metadata(
                         annotation_type = "refseq"
                     else:
                         logger.critical(
-                            f"Unable to find annotation for {strain}." " Skipping..."
+                            f"Unable to find annotation for {strain}. Skipping..."
                         )
                         continue
                 elif annotation_type == "refseq":
@@ -981,7 +1008,7 @@ def extract_metadata(
                         annotation_type = "genbank"
                     else:
                         logger.critical(
-                            f"Unable to find annotation for {strain}." " Skipping..."
+                            f"Unable to find annotation for {strain}. Skipping..."
                         )
                         continue
 
@@ -1001,11 +1028,19 @@ def extract_metadata(
             try:
                 accession = synonym[annotation_type]
             except KeyError:
-                accession = ""
-            else:
-                acc_type += annotation_type
+                accession = ''
             if not accession:
-                logger.warning(f"Unable to find accession for {uid}. " "Skipping...")
+                if annotation_type == "refseq":
+                    annotation_type = "genbank"
+                elif annotation_type == "genbank":
+                    annotation_type = "refseq"
+            try:
+                accession = synonym[annotation_type]
+            except KeyError:
+                accession = ''
+            acc_type += annotation_type
+            if not accession:
+                logger.warning(f"Unable to find accession for {uid}. Skipping...")
                 continue
             else:
                 line.append(annotation_type)
@@ -1042,6 +1077,7 @@ def extract_metadata(
                 "prefix": prefix,
                 "assem_name": assem_name,
                 "swap": False,
+                "biosample": biosample,
             }
             metadata.append(line)
     with open(metadata_file, open_mode) as metadatafh:
@@ -1051,6 +1087,72 @@ def extract_metadata(
     return dl_mapping
 
 
+def get_biosample_data(efetch, xml2json, bs2assemid, metadata_append: bool = False):
+    if metadata_append:
+        open_mode = "at"
+    else:
+        open_mode = "wt"
+    # efetch -db biosample -mode xml | xml2json > biosamples.json
+    logger = logging.getLogger(__name__)
+    command = [f"{efetch}", "-db", "biosample", "-mode", "xml", "|", f"{xml2json}"]
+    try:
+        output = subprocess.run(
+            " ".join(command),
+            stdout=subprocess.PIPE,
+            shell=True,
+            text=True,
+            input="\n".join(sorted(bs2assemid.keys())),
+            check=True,
+        )
+    except subprocess.CalledProcessError as e:
+        logger.critical("There was an error running efetch.")
+        logger.critical(f'Check the error: "{e}" and try again.')
+        exit(1)
+    bsjson = json.loads(output.stdout)["BioSampleSet"]["BioSample"]
+    try:
+        with open(".biosamplejson", open_mode) as bsjsonfh:
+            bsjsonfh.write(output.stdout)
+    except UnboundLocalError:
+        pass
+    biosamples = []
+    for uid in bsjson.keys():
+        if isinstance(bsjson, list):
+            acc = bsjson[uid]["accession"]
+            attrs = bsjson[uid]["Attributes"]["Attribute"]
+        elif isinstance(bsjson, dict):
+            acc = bsjson["accession"]
+            attrs = bsjson["Attributes"]["Attribute"]
+        else:
+            logger.warning(f"Unable to find accession for {uid}. Skipping...")
+            continue
+        bs = BioSample(acc, bs2assemid[acc])
+        if isinstance(attrs, list):
+            for attr in attrs:
+                extract_attribute(bs, attr)
+        elif isinstance(attrs, dict):
+            extract_attribute(bs, attrs)
+        biosamples.append(bs)
+    write_header = True
+    if metadata_append and os.path.exists("biosample.tab"):
+        write_header = False
+    with open("biosample.tab", open_mode) as bsfh:
+        header = ["biosample_acc", "assemid", "isolate"]
+        header.extend(KEYS)
+        if write_header:
+            bsfh.write("\t".join(header))
+            bsfh.write("\n")
+        for bs in sorted(biosamples, key=lambda x: x.acc):
+            line = [
+                bs.acc,
+                bs.assemid,
+                bs.isolate,
+            ]
+            for attr in KEYS:
+                line.append(getattr(bs, attr, "-"))
+            bsfh.write("\t".join(line))
+            bsfh.write("\n")
+
+
 def download_genomes(o, dl_mapping, threads):
     logger = logging.getLogger(__name__)
     filemap = {
@@ -1120,6 +1222,8 @@ def download_genomes(o, dl_mapping, threads):
 
 def main():
     args = run_argparse()
+    logger = logging.getLogger(__name__)
+    logger.info(f'Welcome to get_assemblies v{__version__}!')
     args, exes = validate_inputs(args)
 
     # assem_ids is a list of NCBI assembly IDs
@@ -1157,6 +1261,9 @@ def main():
             docsums,
             args.keepmulti,
         )
+        if not args.biosample:
+            bs2assemid = get_bs2assemid(dl_mapping)
+            get_biosample_data(exes["efetch"], exes["xml2json"], bs2assemid, args.metadata_append)
 
     if "genomes" in args.function:
         download_genomes(args.o, dl_mapping, args.threads)
